%code top{
  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h>
  #include "parser.tab.h"

  extern int yylex();
  extern int yyerror();

  typedef struct car_entry{
    const char *name;
    speed_entry *speed_list;
    struct car_entry *next;
  }car_entry;

  car_entry *car_table = NULL;

  int speed_list_get(speed_entry *list, const char *name){
    speed_entry *s;
    for(s = list; s != NULL; s = s->next){
      if(!strcmp(name, s->name))
        return s->value;
    }
    return -1;
  }

  speed_entry *speed_list_put(speed_entry *list, const char *name, int value){
    if(speed_list_get(list, name) != -1)
      return NULL;
    speed_entry *new_node = (speed_entry *) malloc(sizeof(speed_entry));
    new_node->name = name;
    new_node->value = value;
    new_node->next = list;
    return new_node;
  }

  void free_speed_list(speed_entry *list){
    speed_entry *n, *s = list;
    while(s != NULL){
      free((void *)s->name);
      n = s->next;
      free((void *)s);
      s = n;
    }
  }

  speed_entry *car_table_get(const char *name){
    car_entry *c;
    for(c = car_table; c != NULL; c = c->next)
      if(!strcmp(name, c->name))
        return c->speed_list;
    return NULL;
  }

  int car_table_put(const char *name, speed_entry *speed_list){
    if(car_table_get(name) != NULL)
      return 1;
    car_entry *c = (car_entry *) malloc(sizeof(car_entry));
    c->name = name;
    c->speed_list = speed_list;
    c->next = car_table;
    car_table = c;
    return 0;
  }

  void free_car_table(){
    car_entry *c = car_table;
    car_entry *n;
    while(c != NULL){
      free((void *)c->name);
      free_speed_list(c->speed_list);
      n = c->next;
      free((void *)c);
      c = n;
    }
  }
}
%code requires {
  typedef struct {
      char *name;
      float value;
    }car_performances;
  typedef struct speed_entry{
    const char *name;
    int value;
    struct speed_entry *next; 
  }speed_entry;
}

%union {
  int int_value;
  float float_value;
  char *string;
  speed_entry *speed_list;
  int speeds[2];
  car_performances *car_performances;
}

%debug

//////////////////////////////////
///// SYMBOLS DECLARATION
/////////////////////////////////

%token TOKEN1 TOKEN2 TOKEN3 M MS MINMAX PART ARROW SEP
        EQ PIPE CM S COL RO RC SO SC;
%token <string> QSTRING;
%token <int_value> UINT;

%type <car_performances> performances;
%type <speeds> section_names
%type <speed_list> speeds;
%type <float_value> drive_stats parts part;
%type <string> NT0 NT1;

%%
//////////////////////////////////
///// GRAMMAR
/////////////////////////////////

prog : header SEP cars SEP race;


/******************/
/* Header section */
/******************/
header : token1_l TOKEN2 S token1_l TOKEN3 S token1_l
         | token1_l TOKEN3 S token1_l TOKEN2 S token1_l
;

token1_l : token1_l TOKEN1 S | /* Empty */;


/****************/
/* Cars section */
/****************/
cars : car car | cars car car;

car : QSTRING SO speeds SC {
	     car_table_put($1, $3);
	}
;

speeds : QSTRING EQ UINT MS 
            {
              $$ = speed_list_put(NULL, $1, $3);
            }
          | speeds CM QSTRING EQ UINT MS 
            {
              $$ = speed_list_put($1, $3, $5);
            }
;

/****************/
/* Race section */
/****************/
race : print_min_max_l performances {
        printf("WINNER: %s %.1f s\n", $2->name, $2->value); free((void *)$2->name); free((void *)$2);
	    }
;

/* Management of PRINT_MIN_MAX function */
print_min_max_l : /* Empty */ | print_min_max_l min_max;

min_max : MINMAX RO QSTRING RC RO section_names RC S
        {
          printf("MIN: %d MAX: %d\n", $6[0], $6[1]);
	      }
;

section_names: QSTRING 
              {
                $$[0] = speed_list_get(car_table_get($<string>-2), $1);
                $$[1] = speed_list_get(car_table_get($<string>-2), $1);
              }
		          | section_names CM QSTRING
              {
                if(speed_list_get(car_table_get($<string>-2), $3) < $1[0]){
                  //new min
                  $$[0] = speed_list_get(car_table_get($<string>-2), $3);
                  $$[1] = $1[1];
                }
                else if($1[1] < speed_list_get(car_table_get($<string>-2), $3)){
                  //new max
                  $$[0] = $1[0];
                  $$[1] = speed_list_get(car_table_get($<string>-2), $3);
                }
                else {
                  // no change in min and max
                  $$[0] = $1[0];
                  $$[1] = $1[1];
                }
              }
;

/* Part regarding performances */
performances: QSTRING { printf("%s\n", $1);} ARROW parts S 
            {
              printf("TOTAL: %.1f s\n", $4);
              // To detect the winner car
              $$ = (car_performances *) malloc(sizeof(car_performances));
              $$->name = $<string>-2;
              $$->value = $4;
		        }
	         | performances QSTRING { printf("%s\n", $2); } ARROW parts S 
           {
              printf("TOTAL: %.1f s\n", $5);
              if($1->value < $5){
                // Current winner is an old car
                $$ = $1;
                free((void *)$2);
              }
              else {
                // Current winner is this car
                $$ = $1;
                $$->name = $2;
                $$->value = $5;
              }
            }
;

// The two markers are used to transfer the car name semantic value in the symbol that preceeds the non terminal "part"
parts : NT0 part { $$ = $2; }
	  | parts PIPE NT1 part { $$ = $1 + $4;}
;

NT0 : { $$ = $<string>-2; };

NT1 : { $$ = $<string>-4; };

part : PART UINT COL drive_stats {
              $$ = $4;
              printf("PART %d: %.1f s\n", $2, $4);
};

drive_stats : QSTRING UINT M 
            {
              $$ = (float)$2 / speed_list_get(car_table_get($<string>-3), $1);
		        }
		        | drive_stats CM QSTRING UINT M 
            {
              $$ = $1 + (float)$4 / speed_list_get(car_table_get($<string>-3), $3);
            }
;